/*
  acpiw.c: basic functions to access /proc/acpi
  (C) 2002 Oliver Diedrich

  This file may be redistributed under the terms of the
  GNU Public License (GPL), see www.gnu.org
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/klog.h>

#include "acpiw.h"

/********************/
/* global variables */
/********************/

int event;     /* file descriptor for /proc/acpi/event, must stay opened */
               /* all the time in order to get all events */




/*************/
/* functions */
/*************/

void usage(char *myname)
{
  printf("usage: %s [-u n] [-s fname]\n", myname);
  printf("-u n: set update interval to n seconds (default: 5; 0: don't update)\n");
  printf("-s fname: save ACPI data in file fname\n");
  printf("\n");
  exit(0);
} /* usage() */




file *new_file(char *name)
/* mallocs a new file structure and copies name to file->name if name != NULL */
/* NULL: error, else: pointer to file structure */
{
  file *f = malloc(sizeof(file));
  if (!f) return NULL;
  f->mode = 0;
  f->data = NULL;
  f->changed = 0;
  f->next = NULL;
  f->iter = NULL;
  if (name)
  {
    f->name = malloc(strlen(name)+1);
    if (!f->name) return NULL;
    memset(f->name, 0, strlen(name)+1);
    strcpy(f->name, name);
  }
  else f->name = NULL;
  return f;
} /* new_file() */





line *new_line(long seconds, char *label, char *value)
/* creates a new line* structure and copies label and value to the respective */
/* fields if != NULL; seconds is put into the time field */
/* NULL: error, else: pointer to file structure */
{
  line *l = malloc(sizeof(line));
  if (!l) return NULL;
  l->time = seconds;
  l->newer = NULL;
  l->latest = l;
  l->changed = 0;
  l->next = NULL;
  l->iter = NULL;
  if (label)
  {
    l->label = malloc(strlen(label)+1);
    if (!l->label) return NULL;
    memset(l->label, 0, strlen(label)+1);
    strcpy(l->label, label);
  }
  else l->label = NULL;
  if (value)
  {
    l->value = malloc(strlen(value)+1);
    if (!l->value) return NULL;
    memset(l->value, 0, strlen(value)+1);
    strcpy(l->value, value);
  }
  else l->value = NULL;
  return l;
} /* new_line() */




line *read_bin_file(char *name, long seconds)
/* reads the content of binary file name and stores it in a line struct */
/* data are stored in line->value from the 9th byte on; byte 0 is ' ', 1-3 */
/* are '\0', bytes 4-7 take the length of the data */
/* seconds contains time in seconds since the epoch */
/* NULL: error, else: line structure */
{
  int f;                 /* the file */
  char buf[512];         /* read buffer */
  line *l;               /* a line entry for the binary data */
  int n_read;            /* bytes just read */
  int size;              /* total number of bytes read + 8 leading Bytes */
  int *p;                /* helps to set the number of read bytes in l */

assert(name != NULL);
assert(name[0] != '\0');

  f = open(name, O_RDONLY);
  if (-1 == f) return NULL;

  l = new_line(seconds, name, NULL);
  if (!l) return NULL;
  l->value = malloc(8);
  memset(l->value, 0, 8);     /* 8 leading bytes: " \0\0\0" + (int *) # bytes */
  size = 8;
  l->value[0] = ' ';     /* "" as l->value indicates a dir/file entry in the */
                         /* gtk tree in gui.c, so we need a valid 1st char. */
                         /* A '\0' after this char avoids binary data being */
                         /* treated as a text value and printed to screen */
  l->time = seconds;

  while ( (n_read = read(f, buf, 512)) > 0)
  {
    l->value = realloc(l->value, size+n_read);
    if (!l->value) return NULL;
    memcpy(l->value+size, buf, n_read);
    size += n_read;
  }
  p = (int *) (l->value+4);
  *p = size;
  close(f);
  return l;
} /* read_bin_file() */




line *readfile(char *name, long seconds)
/* reads the content of file name and stores it in a list of line struct's */
/* seconds contains time in seconds since the epoch */
/* NULL: error, else: start of list */
{
  FILE *f;
  char str[256];         /* read buffer */
  line *st = NULL;       /* start of line list */
  line *last = NULL;     /* last file entry read */
  line *l;               /* recent file entry in list */
  int len;               /* length of string read from f */
  char *value_start;     /* start of value in line */
  char *c;

assert(name != NULL);
assert(name[0] != '\0');

  f = fopen(name, "r");
  if (!f) return NULL;

  while (fgets(str, 255, f))  /* read all lines in the file */
  {
    len = strlen(str);
    if (!len) continue;       /* empty string */
    if ('\n' == str[len-1]) str[len-1] = '\0';
    l = new_line(seconds, NULL, NULL);  /* create a new line entry */

    c = str+1;
    while (*c > '\0')
    {
      if ((' ' == *c) && (':' == *(c-1)))    /* end of label */
      {
        l->label = malloc(c-str+1);
        if (!l->label) return NULL;
        strncpy(l->label, str, c-str);
        l->label[c-str] = '\0';
        while ((' ' == *c) && (*c > '\0')) c++;
        if ('\0' == *c)       /* end of string: label, but no value */
        {
          l->value = malloc(1);
          l->value[0] = '\0';
          value_start = 0;
        }
        else
        {
          value_start = c;
          while (*c > '\0') c++;
        }
      }
      if (*c) c++;       /* else: string has ended after label */
    } /* while (*c) */

    if (l->label)        /* label has been found */
    {
      if (value_start)   /* value exists */
      {
        l->value = malloc(c-value_start+1);
        if (!l->value) return NULL;
        strcpy(l->value, value_start);
      }
    }
    else                 /* only label without value and ": " */
    {
      l->label = malloc(strlen(str)+1);
      strcpy(l->label, str);
    }

    if (!st) st = l;
    if (last) last->next = l;
    last = l;
  }
  fclose(f);
  return st;
} /* readfile() */    




file *build_tree(char *dirname)
/* scans directory dirname recursively and builds a directory and file tree */
/* NULL: error, else: start of tree (1st file in dirname) */
{
  DIR *dir;              /* dir handle to scan directory */
  struct dirent *dent;   /* directory entry */
  struct stat buf;       /* needed for fstat(dent) */
  file *f;               /* file entry in internal tree */
  file *last = NULL;     /* last entry before current entry */
  file *root = NULL;     /* start of tree for dirname */
  char path[255];        /* dirname + name of file/subdir */
  line *l;               /* to store error messages instead of file content */
  time_t seconds;        /* time in seconds after the epoch */
  struct timeval tv;     /* needed for gettimeofday() */

  if (-1 == gettimeofday(&tv, NULL)) seconds = 0;
  else seconds = tv.tv_sec;

  dir = opendir(dirname);
  if (!dir) return NULL;

  while ( (dent = readdir(dir)) )  /* read all directory entries */
  {
    if (!strncmp(".", dent->d_name, 1)) continue;
    if (!strncmp("..", dent->d_name, 2)) continue;

    f = new_file(dent->d_name);
    sprintf(path, "%s/%s", dirname, f->name);
    if (-1 == stat(path, &buf)) return NULL;
    f->mode = buf.st_mode;

    if (!root) root = f;      /* first file in directory */
    if (last) last->next = f;

    if (S_ISDIR(f->mode)) f->data = build_tree(path);
    else                      /* regular file */
    {
      if (!strncmp("dt", &f->name[strlen(f->name)-2], 2))   /* "*dt" files */
      {                                             /* contain binary data */
        f->data = read_bin_file(path, seconds);
        if (!f->data)         /* read error */
        {
          l = new_line(seconds, "### read error ###", NULL);
          f->data = l;
        }
      }
      else if (!strcmp(f->name, "event"))    /* /proc/acpi/event */
      {                  /* stays permanently opened to get all events */
        event = open(path, O_RDONLY | O_NONBLOCK);
        if (-1 == event)
        {
          l = new_line(seconds, "### read error ###", NULL);
          if (!l) return NULL;
          f->data = l;
        }
        else f->data = NULL;       /* nothing read yet */
      }
      else                         /* "normal" file: read the file content */
      {
        f->data = readfile(path, seconds);
        if (!f->data)              /* read error */
        {
          l = new_line(seconds, "### read error ###", NULL);
          if (!l) return NULL;
          f->data = l;
        }
      }
    } /* else: regular file */
    last = f;
  } /* while */
  closedir(dir);
  return root;
} /* build_tree() */

    


int rescan_tree(file *tree, char *basedir)
/* scans tree recursively for changes by reading files in basedir */
/* 1: change occured, 0: no change */
{
  file *f;               /* walks along tree */
  line *old;             /* old file content */
  line *new;             /* recent file content */
  line *tmp;             /* to create new struct line */
  char path[256];        /* basedir + recent file name */
  int changed = 0;       /* 1: something in tree has changed */
  time_t seconds;        /* time in seconds after the epoch */
  struct timeval tv;     /* needed for gettimeofday() */

assert(tree != NULL);
assert(basedir != NULL);

  if (-1 == gettimeofday(&tv, NULL)) seconds = 0;
  else seconds = tv.tv_sec;

  f = tree;
  while(f)
  {
    f->changed = 0;                     /* reset changed flag to unchanged */

    if ((S_ISDIR(f->mode) && f->data))
    {
      sprintf(path, "%s/%s", basedir, f->name);
      if (rescan_tree(f->data, path))
      {
        f->changed = 1;
        changed = 1;
      }
    }

    else                                /* regular file */
    {
      old = f->data;                    /* old file content */

      if (!strcmp(f->name, "event"))    /* /proc/acpi/event */
      {
        if (event != -1)
        {
          char buf[256];
          memset(buf, 0, 256);
          while (read(event, buf, 255) > 0)  /* read all ACPI events that */
          {                             /* happened since the last update */
            buf[strlen(buf)-1] = '\0';
            if (!old)                        /* 1st event */
            {
              old = new_line(seconds, buf, NULL);
              f->data = old;
            }
            else
            {
              old->latest->newer = new_line(seconds, buf, NULL);
              old->latest = old->latest->newer;
            }
            old->changed = 1;   /* mark this line as changed */
            f->changed = 1;     /* mark the file as changed */
            changed = 1;        /* set the changed flag for this subtree */
            memset(buf, 0, 256);
          } /* while */
        }
      } /* if ("event") */

      else if (old && strncmp("###", old->label, 3)   /* file with content */
                   && strncmp("dt", &f->name[strlen(f->name)-2], 2))
      {
        sprintf(path, "%s/%s", basedir, f->name);
        new = readfile(path, seconds);

assert(new != NULL);     /* there is no reason why readfile() should fail */
                           /* now, so we just skip the error handling ... */
        while(old)       /* run across all lines in the file */
        {

assert(new != NULL);     /* !!! we assume that the number of lines in a file */
                             /* does not change. Is this always correct? !!! */

          old->changed = 0;   /* reset changed flag */
          if (strcmp(new->value ? new->value : new->label, 
                     new->value ? old->latest->value : old->latest->label))
          {    /* line just read differs from the one read last time */
            f->changed = 1;        /* mark file as changed */
            changed = 1;           /* return "subtree has changed" */
            old->changed = 1;      /* mark entry as changed */

            tmp = new_line(seconds, new->label, new->value);

assert(tmp != NULL);     /* no better error handling possible here */

            tmp->next = old->next; /* next line in file is the same for newer */
                                   /* versions of the recent line */
            old->latest->newer = tmp;
            old->latest = tmp;
          }
          tmp = new;
          old = old->next;
          new = new->next;
          if (tmp->label) free(tmp->label);
          if (tmp->value) free(tmp->value);
          free(tmp);
        } /* while(old) */

assert(new == NULL);     /* we assume that the number of lines in a file */
                         /* does not change */
      } /* if (file with content); else: do nothing with this file */
    } /* else (regular file) */

    f = f->next;
  } /* while(f) */
  return changed;
} /* rescan_tree() */




void dump(unsigned char *buf, FILE *f)
/* dumps the binary data in buf to text. Attention: Bytes 0-3 in buf are */
/* garbage, bytes 4-7 contain the length of the buffer including the eight */
/* leading bytes */
{
  int *size = (int *) (buf+4);     /* size of buf is stored at offset 4 */
  char str[80];                    /* string buffer */
  char *c = str;
  int n = 0;                       /* bytes written */
  int i, j;

  buf += 8;
  *size -= 8;

  for (i=0; i<((*size)/16)+1; i++)
  {
    c = str;
    sprintf(c, "%05X  ", i*16);
    c += 7;
    for (j=0; j<16; j++, buf++, n++)    /* hex display */
    {
      if (8 == j)
      {
        sprintf(c, " ");
        c++;
      }
      if (n<*size) sprintf(c, "%02X ", *buf);
      else sprintf(c, "   ");
      c += 3;
    }
    buf -= j;
    n -= j;
    sprintf(c, " ");
    c++;
    for (j=0; j<16; j++, buf++, n++)    /* character display */
    {
      if (8 == j)
      {
        sprintf(c, " ");
        c++;
      }
      if (n<*size) sprintf(c, "%c", ((*buf < 32) || (*buf > 127)) ? '.' : *buf);
      c++;
    }
    fprintf(f, "%s\n", str);
  } /* for i */
} /* dump() */




int save_tree(file *tree, FILE *f, char *basedir)
/* saves (sub)tree tree in file f. basedir is added to the file names in tree */
/* 0: error, else: ok */
{
  char path[256];

  while (tree)           /* across all files in the subtree */
  {
    if ((S_ISDIR(tree->mode) && tree->data))
    {
      sprintf(path, "%s/%s", basedir, tree->name);
      if (!save_tree(tree->data, f, path)) return 0;
    }

    else
    {
      struct tm tim;
      line *l = tree->data;
      fprintf(f, "\n%s/%s\n", basedir, tree->name);
      if (!strncmp("dt", &tree->name[strlen(tree->name)-2], 2))
      {
        if (l->value) dump(l->value, f);
        else
        {
          localtime_r(&(l->time), &tim);
          if (tim.tm_year > 100) tim.tm_year -= 100;
          fprintf(f, "(%2d/%02d/%02d %2d:%02d:%02d)\t", tim.tm_mon+1,
                         tim.tm_mday, tim.tm_year, tim.tm_hour, tim.tm_min,
                         tim.tm_sec);
          fprintf(f, "### read error ###\n");
        }
      }
      else
      {
        while (l)          /* across all lines of the file */
        {
          if (l->label)
          {
            line *l1 = l;
            while (l1)     /* across all newer version of this line */
            {
              localtime_r(&(l1->time), &tim);
              if (tim.tm_year > 100) tim.tm_year -= 100;
              fprintf(f, "(%2d/%02d/%02d %2d:%02d:%02d)\t%s\t%s\n", tim.tm_mon+1,
                         tim.tm_mday, tim.tm_year, tim.tm_hour, tim.tm_min,
                         tim.tm_sec, l1->label, l1->value ? l1->value : "");
              l1 = l1->newer;
            }
          }
          l = l->next;
        } /* while (l) */
      }
    }
    tree = tree->next;
  }
  return 1;
} /* save_tree() */




int save_all(file *acpi_tree, FILE *f, char *basedir, char **acpi_info, 
             int n_acpi_info, char acpi_version[], sysinfo si)
/* writes ACPI syslog info and ACPI tree in file f */
/* 0: error, 1: ok */
{
  int i;
  char **c = acpi_info;

  fprintf(f, "%s\n(C) 2002 Oliver Diedrich\n\n", VERSION);

  fprintf(f, "System information\n\n");
  fprintf(f, "System:\t%s %s (%s); serial #: %s\n", 
             si.sys_vendor, si.sys_product,
             si.chassis, si.sys_serial);
  fprintf(f, "BIOS:\t%s %s (%s)\n",
             si.bios_vendor, si.bios_version, si.bios_release);
  fprintf(f, "Board:\t%s %s (%s); serial #: %s\n",
              si.board_vendor, si.board_product,
              si.board_version, si.board_serial);
  fprintf(f, "CPU:\t%s %s, %d MHz\n", si.cpu_maker, si.cpu, si.mhz);
  fprintf(f, "RAM:\t%d MByte\n", si.mem / (1024*1023));
  fprintf(f, "Kernel:\t%s %s for %s (build %s)\n",
             si.sysname, si.release, si.machine, si.version);
  fprintf(f, "ACPI:\trelease %s\n", acpi_version ? acpi_version : "");

  fprintf(f, "\nACPI related syslog messages:\n\n");
  for (i=0; i<n_acpi_info; i++, c++)
  {
    fprintf(f, "%s\n", *c);
  }
  fprintf(f, "\n");

  if (save_tree(acpi_tree, f, PROC)) return 1;
  else return 0;
} /* save_all() */





int get_acpi_version(char *basedir, char *version)
/* reads ACPI version from basedir/info->version: and copies it to version */
/* buf must be large enough (10 chars should be enough) */
/*  0: no version string found, 1: ok */
{
  line *l;
  char fname[256];

  sprintf(fname, "%s/info", basedir);
  l = readfile(fname, 0);
  while (l)
  {
    if (l->label && (!strncmp("version", l->label, 7)) && l->value)
    {
      sprintf(version, l->value);
      return 1;
    }
    l = l->next;
  }
  return 0;
} /* get_acpi_version() */




int get_acpi_info(char ***acpi_info)
/* reads ACPI related info from the syslog and writes as an array of char* to */
/* acpi_info. all necessary memory is allocated */
/* 0: error or nothing found, else: number of strings in buf */
{
  char log_buf[16384];
  int n, i, n_lines;
  char *c;
  char *str;
  char **buf, **save;

  memset(log_buf, 0, 16384);
  n = klogctl(3, log_buf, 16384);
  c = log_buf;
  n_lines = 0;
  for (i=0; i<n; i++,c++)
  {
    if ('\n' == *c)
    {
      *c = '\0';
      n_lines++;
    }
  }

  buf = malloc(n_lines*sizeof(char *));
  if (!buf) return 0;
  save = buf;
  c = log_buf;
  n = 0;
  for (i=0; i<n_lines; i++)
  {
    c += 3;              /* syslog entries start with <n> */
    if (!strncmp(c, "ACPI", 4))
    {
      str = malloc(strlen(c)+1);
      if (!str) return 0;
      memset(str, 0, strlen(c)+1);
      strcpy(str, c);
      *buf = str;
      buf++;
      n++;
    }
    else
    {
      while (*c > '\0') c++;
      c++;
    }
  }
  *acpi_info = save;
  return n;
} /* get_acpi_info() */
