/*
   gui.c: GUI for acpi watch tool acpiw
  (C) 2002 Oliver Diedrich

  This file may be redistributed under the terms of the
  GNU Public License (GPL), see www.gnu.org
*/


#include <gtk/gtk.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#include "acpiw.h"
#include "menu.h"



/* the Gtk tree model contains 4 columns for each entry: */
/* 0: name of dir/file or label of file entry */
/* 1: value of file entry ("" on dir/file names, " " on file  entries with */
/*    a label, but without value) */
/* 2: color as string (red: changed; blue: file that can be written to; else: */
/*    black */
/* 3: pointer to data in our ACPI tree (entry* or line*, can be distinguished */
/*    by strlen(value)) */

#define COL_LABEL 0
#define COL_VALUE 1
#define COL_COLOR 2
#define COL_PTR   3
#define COL_TYPE  4

/* what can be stored in COL_TYPE */

#define TYPE_INVALID 0
#define TYPE_FILE    1
#define TYPE_LINE    2


/********************/
/* global variables */
/********************/

line *selected = NULL;        /* entry just selected to watch in text window */
                              /* needed to update the text window */
char *fname = NULL;           /* file to write ACPI data to; needed in */
                              /* handler for "save" menu item */
file *acpi_tree;              /* internal ACPI tree from /proc/acpi */
GtkTreeStore *gtk_tree;       /* internal Gtk model of ACPI tree */
GtkWidget *gtk_tree_view;     /* the Gtk tree view of the ACPI tree */
GtkTextBuffer *text;          /* text to display in text_view */
char **acpi_info;             /* ACPI info from the syslog */
int n_acpi_info;              /* # strings in acpi_info */
char acpi_version[10];        /* /proc/acpi/info->version: */
sysinfo si;                   /* system information */
volatile int lock = 0;        /* 1 while in update_handler */
int debug = 0;                /* 1: print debug info in update_handler() (-v) */



void  hexdump(unsigned char *buf, GtkTextBuffer *text, GtkTextIter* it)
/* dumps the binary data in buf to text. Attention: Bytes 0-3 in buf are */
/* garbage, bytes 4-7 contain the length of the buffer including the eight */
/* leading bytes */
{
  int *size = (int *) (buf+4);     /* size of buf is stored at offset 4 */
  char str[80];                    /* string buffer */
  char *c = str;
  int n = 0;                       /* bytes written */
  int i, j;

  buf += 8;
  *size -= 8;

  for (i=0; i<((*size)/16)+1; i++)
  {
    c = str;
    sprintf(c, "%05X  ", i*16);
    c += 7;
    for (j=0; j<16; j++, buf++, n++)    /* hex display */
    {
      if (8 == j)
      {
        sprintf(c, " ");
        c++;
      }
      if (n<*size) sprintf(c, "%02X ", *buf);
      else sprintf(c, "   ");
      c += 3;
    }
    buf -= j;
    n -= j;
    sprintf(c, " ");
    c++;
    for (j=0; j<16; j++, buf++, n++)    /* character display */
    {
      if (8 == j)
      {
        sprintf(c, " ");
        c++;
      }
      if (n<*size) sprintf(c, "%c", ((*buf < 32) || (*buf > 127)) ? '.' : *buf);
      c++;
    }
    sprintf(c, "\n");
    gtk_text_buffer_insert_with_tags_by_name(text, it, str, -1, "monospace", 0);
  } /* for i */
} /* hexdump() */




void on_selection_changed(GtkTreeSelection *select, GtkTextBuffer *text)
/* called if selection in the tree changes. text points to the text buffer */
/* associated with the text view on the right side of the window */
/* displays history of entries in lines or access rights of files */
{
  GtkTreeIter iter;      /* selected entry in selection */
  GtkTreeModel *model;   /* the tree model associated with select */
  GtkTextIter it;        /* needed to write to text */
  void *p;               /* ACPI tree item pointer from selected entry */
  struct tm tim;
  char str[1024];
  char *c;
  int type;
  
  if (gtk_tree_selection_get_selected(select, &model, &iter))
  {
    gtk_tree_model_get(model, &iter, COL_VALUE, &c, COL_PTR, &p,
                       COL_TYPE, &type, -1);

    if (TYPE_FILE == type)    /* file or directory -> p pointes to an entry */
    {
      file *f = p;
      selected = NULL;   /* only line entries are updated when selected */
      gtk_text_buffer_set_text(text, "\0", -1);   /* delete text buffer */
      gtk_text_buffer_get_iter_at_offset(text, &it, 0);
      sprintf(str, "%c%c%c%c%c%c%c%c%c%c  ",      /* access rights */
                   S_ISDIR(f->mode) ? 'd' : '-',
                   (S_IRUSR & f->mode) ? 'r' : '-',
                   (S_IWUSR & f->mode) ? 'w' : '-',
                   (S_IXUSR & f->mode) ? 'x' : '-',
                   (S_IRGRP & f->mode) ? 'r' : '-',
                   (S_IWGRP & f->mode) ? 'w' : '-',
                   (S_IXGRP & f->mode) ? 'x' : '-',
                   (S_IROTH & f->mode) ? 'r' : '-',
                   (S_IWOTH & f->mode) ? 'w' : '-',
                   (S_IXOTH & f->mode) ? 'x' : '-');
      gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, "monospace",
                                               0);
      sprintf(str, "%s\n", f->name);
      gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, 0);
    }

    else                 /* a file entry -> p points to a line */
    {
      line *l = p;
      selected = l;
      gtk_text_buffer_set_text(text, "\0", -1);
      gtk_text_buffer_get_iter_at_offset(text, &it, 0);

      if (l->value)
      {
        sprintf(str, "%s\n", l->label);
        gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, "bold", 0);
        if (!strncmp(l->label, "/proc/", 6))      /* a *dt file */
         hexdump(l->value, text, &it);
        else
        {
          while (l)
          {
            localtime_r(&(l->time), &tim);
            sprintf(str, "%2d/%02d %02d:%02d:%02d\t ", tim.tm_mon+1, tim.tm_mday,
                          tim.tm_hour, tim.tm_min, tim.tm_sec);
            gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, 0);
            sprintf(str, "%s\n", l->value);
            gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1,
                                                     "monospace", 0);
            l = l->newer;
          }
        }
      } /* if (l->value) */
      else
      {
        while (l)
        {
          localtime_r(&(l->time), &tim);
          sprintf(str, "%2d/%02d %02d:%02d:%02d\t ", tim.tm_mon+1, tim.tm_mday,
                        tim.tm_hour, tim.tm_min, tim.tm_sec);
          gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, 0);
          sprintf(str, "%s\n", l->label);
          gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, 
                                                   "monospace", 0);
          l = l->newer;
        }
      }
    } /* else */

  } /* if (gtk_tree_selection_get_selected()) */
} /* on_selection_changed() */




void on_activated(GtkTreeView *tree_view, GtkTreePath *path,
                  GtkTreeViewColumn *column, gpointer nil)
/* called if item path in the tree tree_view is double clicked */
/* expands or collects the subtree path */
{
  GtkTreeModel *model = gtk_tree_view_get_model(tree_view);
  GtkTreeIter iter;
  void *p;
  char *c;
  int type;

  if (gtk_tree_model_get_iter(model, &iter, path))
  {
    gtk_tree_model_get(model, &iter, COL_VALUE, &c, COL_PTR, &p,
                       COL_TYPE, &type, -1);
    if (TYPE_FILE == type)    /* file or directory */
    {
      if (gtk_tree_view_row_expanded(tree_view, path))
       gtk_tree_view_collapse_row(tree_view, path);
      else gtk_tree_view_expand_row(tree_view, path, TRUE);
    }
  }
} /* on_activated() */




void update_gtk_tree(file *tree, GtkTreeStore *gtk_tree)
/* updates the content of the gtk_tree tree if something in our ACPI tree */
/* has changed. Will be called usually after a rescan_tree() */
{
  while(tree)
  {
    if (tree->changed)   /* if a directory or file is not marked as changed, */
    {                    /* nothing of its content has changed, and no action */
                         /* is necessary */
      gtk_tree_store_set(gtk_tree, tree->iter, COL_COLOR, "red", -1);
      if (S_ISDIR(tree->mode) && tree->data) 
       update_gtk_tree(tree->data, gtk_tree);
      else
      { 
        line *l = tree->data;
        while(l)
        {
          if (l->changed)
          {
            if (l->iter)
            {
              gtk_tree_store_set(gtk_tree, l->iter, 
                                 COL_LABEL, l->latest->label,
                                 COL_VALUE, l->latest->value ? 
                                            l->latest->value : " ",
                                 COL_COLOR, "red",
                                 -1);
            }
            else    /* 1st entry in a previously empty file */
            {
              GtkTreeIter *iter = malloc(sizeof(GtkTreeIter));
              memset(iter, 0, sizeof(GtkTreeIter));
              gtk_tree_store_insert_before(gtk_tree, iter, tree->iter, NULL);
              gtk_tree_store_set(gtk_tree, iter,
                                 COL_LABEL, l->label,
                                 COL_VALUE, l->value ? l->value : "",
                                 COL_COLOR, "red",
                                 COL_PTR, l,
                                 COL_TYPE, TYPE_LINE, -1);
              l->iter = iter;
            }
          } /* if (l->changed) */
          l = l->next;
        }
      }
    }
    tree = tree->next;
  } /* while(tree) */
} /* update_gtk_tree() */




void update_handler(int x)
/* called each update seconds from gtk_main() */
/* vars[] contains pointers to the internal ACPI tree, the gtk tree, */
/* the gtk tree view, the tree selection, and the text buffer */
/* calls rescan_tree(); if there are any changes, the gtk tree is updated; */
/* the tree view is redrawn; if a file entry is displayed on the right and */
/* has changed, its history is updated */
{
  time_t seconds;
  struct timeval tv;

  if (lock)
  {
    puts(">>> update_handler() called too fast, pausing\n");
    return;
  }

  lock = 1;
  if (debug) puts(">>> update_handler() started");
  if (-1 == gettimeofday(&tv, NULL)) seconds = 0;      /* get recent time */
  else seconds = tv.tv_sec;
  if (debug>1) puts(">>> starting rescan_tree()");
  if (rescan_tree(acpi_tree, PROC))     /* there are changes in the ACPI tree */
  {
    if (debug>1) puts(">>> rescan_tree() finished");
    if (debug>1) puts(">>> starting update_gtk_tree()");
    update_gtk_tree(acpi_tree, gtk_tree);
    if (debug>1) puts(">>> update_gtk_tree() finished");
    if (debug>1)  puts(">>> starting gtk_tree_view_set_model()");
    gtk_tree_view_set_model(GTK_TREE_VIEW(gtk_tree_view),
                            GTK_TREE_MODEL(gtk_tree));
    if (debug>1) puts(">>> gtk_tree_view_set_model finished()");
    if ((selected) && (selected->latest->time-seconds >= 0))     /* the entry */
    {                                 /* watched in the text view has changed */
      GtkTextIter it;
      struct tm tim;
      char str[1024];
      line *l = selected->latest;

      if (debug>1) puts(">>> updating text buffer");
      gtk_text_buffer_get_end_iter(text, &it);
      localtime_r(&(l->time), &tim);
      sprintf(str, "%2d/%02d %02d:%02d:%02d\t ", tim.tm_mon+1, tim.tm_mday,
                   tim.tm_hour, tim.tm_min, tim.tm_sec);
      gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, 0);
      if (l->value) sprintf(str, "%s\n", l->value);
      else sprintf(str, "%s\n", l->label);
      gtk_text_buffer_insert_with_tags_by_name(text, &it, str, -1, 
                                               "monospace", 0);
      if (debug>1) puts(">>> text buffer updated");
    }
  }
  if (debug) puts(">>> update_handler() finished\n");
  lock = 0;
} /* update_handler() */





void build_gtk_tree(GtkTreeStore *gtk_tree, GtkTreeIter *parent, file *tree)
/* puts (sub)tree tree into GtkTreeStore gtk_tree under GtkTreeIter parent */
{
  GtkTreeIter *iter;
  GtkTreeIter *l_iter;

  while (tree)
  {
    iter = malloc(sizeof(GtkTreeIter));
    memset(iter, 0, sizeof(GtkTreeIter));
    gtk_tree_store_insert_before(gtk_tree, iter, parent, NULL);
    if (S_ISREG(tree->mode) && (S_IWUSR & tree->mode)) /* writeable files ... */
    {
      char str[80];
      sprintf(str, "<tt><small>%c%c%c%c%c%c%c%c%c%c</small></tt>",      /* access rights */
                   S_ISDIR(tree->mode) ? 'd' : '-',
                   (S_IRUSR & tree->mode) ? 'r' : '-',
                   (S_IWUSR & tree->mode) ? 'w' : '-',
                   (S_IXUSR & tree->mode) ? 'x' : '-',
                   (S_IRGRP & tree->mode) ? 'r' : '-',
                   (S_IWGRP & tree->mode) ? 'w' : '-',
                   (S_IXGRP & tree->mode) ? 'x' : '-',
                   (S_IROTH & tree->mode) ? 'r' : '-',
                   (S_IWOTH & tree->mode) ? 'w' : '-',
                   (S_IXOTH & tree->mode) ? 'x' : '-');
      gtk_tree_store_set(gtk_tree, iter, COL_LABEL, tree->name,
                                         COL_VALUE, str, 
                                         COL_COLOR, "blue",    /* ... in blue */
                                         COL_PTR, tree,
                                         COL_TYPE, TYPE_FILE, -1);
    }
    else
    {
      char str[80];
      sprintf(str, "<tt><small>%c%c%c%c%c%c%c%c%c%c</small></tt>",      /* access rights */
                   S_ISDIR(tree->mode) ? 'd' : '-',
                   (S_IRUSR & tree->mode) ? 'r' : '-',
                   (S_IWUSR & tree->mode) ? 'w' : '-',
                   (S_IXUSR & tree->mode) ? 'x' : '-',
                   (S_IRGRP & tree->mode) ? 'r' : '-',
                   (S_IWGRP & tree->mode) ? 'w' : '-',
                   (S_IXGRP & tree->mode) ? 'x' : '-',
                   (S_IROTH & tree->mode) ? 'r' : '-',
                   (S_IWOTH & tree->mode) ? 'w' : '-',
                   (S_IXOTH & tree->mode) ? 'x' : '-');
      gtk_tree_store_set(gtk_tree, iter, COL_LABEL, tree->name, 
                                         COL_VALUE, str, 
                                         COL_COLOR, "black",
                                         COL_PTR, tree,
                                         COL_TYPE, TYPE_FILE, -1);
    }
    tree->iter = iter;

    if (S_ISDIR(tree->mode) && tree->data)
     build_gtk_tree(gtk_tree, iter, tree->data);
    else       /* tree is a file, so tree->data points to its content */
    {
      if (tree->data)
      {
        char str[1024];
        char *c1, *c2;
        line *l = tree->data;
        while(l)
        {
          l_iter = malloc(sizeof(GtkTreeIter));
          memset(l_iter, 0, sizeof(GtkTreeIter));
          gtk_tree_store_insert_before(gtk_tree, l_iter, iter, NULL);
          memset(str, 0, 1024);
          c1 = l->value;
          c2 = str;
          while (c1 && (*c1 > '\0'))
          {
            if ('<' == *c1) c2 += sprintf(c2, "&lt;");
            else if ('>' == *c1) c2 += sprintf(c2, "&gt;");
            else
            {
              *c2 = *c1;
              c2++;
            }
            c1++;
          }
          gtk_tree_store_set(gtk_tree, l_iter,
                             COL_LABEL, l->label,
                             COL_VALUE, l->value ? str : "",
                             COL_COLOR, "black",
                             COL_PTR, l,
                             COL_TYPE, TYPE_LINE, -1);
          l->iter = l_iter;
          l = l->next;
        }
      }
    }
    tree = tree->next;
  } /* while(tree) */
} /* build_gtk_tree() */




int main(int argc, char *argv[])
{
  GtkWidget *window,          /* main window */
            *vbox,            /* vertical container: menu, hbox, status line */
            *menubar,         /* menu */
            *pane,            /* horizontally adjustables panes: tree, text */
            *scrolled_win1,   /* window with scroll bars for the tree */
            *scrolled_win2,   /* window with scroll bars for text */
            *text_view,       /* displays text stored in text */
            *statusbar;       /* status line at bottom of window */
  GtkAccelGroup *accel;       /* accelerator keys */
  guint status_id;            /* for statusbar operations */

  GtkTreeViewColumn *column1; /* a display column in the tree view (labels */
                              /* and directory/file names) */
  GtkTreeViewColumn *column2; /* a display column in the tree view (values) */
  GtkCellRenderer *renderer1, /* text renderer for the column's cells content */
                  *renderer2;
  GtkTreeSelection *select;   /* a selection within tree view */
  GtkTextIter iter;           /* needed for text operations on text buffer */

  int update = 5;             /* tree udate interval */
  int writeable = 0;          /* 1: text buffer writeable (-w) */
  int sig = 1;                /* 0: use gtk_timeout_add() for updates (-g) */
  struct itimerval timer;     /* generates ALRMs each update seconds if sig */
  char str[256];
  int i;

  for (i=1; i<argc; i++)
  {
    char *c = argv[i];
    if (*c != '-') usage(argv[0]);
    c++;
    switch(*c)
    {
      case 'u': if (i==argc-1) usage(argv[0]);    /* last argument */
                i++;
                update = strtol(argv[i], NULL, 10);
                break;

      case 's': if (i==argc-1) usage(argv[0]);    /* last argument */
                i++;
                fname = argv[i];
                break;

      case 'v': while ('v' == *c)
                {
                  debug++;
                  c++;
                }
                break;

      case 'w': writeable = 1;
                break;

      case 'g': sig = 0;
                break;

      default:  usage(argv[0]);
    }
  }

  gtk_init (&argc, &argv);

  get_basic_sysinfo(&si);
  get_dmi_info(&si);


/* create internal /proc/acpi tree acpi_tree */
  if (!(acpi_tree = build_tree(PROC)))
  {
    fprintf(stderr, "could not read %s\n", PROC);
    return 1;
  }
  memset(acpi_version, 0, 10);
  get_acpi_version(PROC, acpi_version);
  n_acpi_info = get_acpi_info(&acpi_info);

/* the Gtk tree model contains 4 columns: 0: name of dir/file or label of    */
/* file entry; 1: value of file entry ("" on dir/file names, " " on file     */
/* entries with a label, but without value); 2: color; 3: pointer to data    */
/* (entry* or line*, can be distinguished by strlen(value)) in our acpi tree */
  gtk_tree = gtk_tree_store_new(5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
                                G_TYPE_POINTER, G_TYPE_INT);
  build_gtk_tree(gtk_tree, NULL, acpi_tree);

/* gtk_tree_view must be given to create_menu(), so we need it here */
/* buffer associated with text_view is needed in tree callback functions */
  gtk_tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(gtk_tree));
  gtk_widget_show(gtk_tree_view);
  select = gtk_tree_view_get_selection(GTK_TREE_VIEW(gtk_tree_view));
  text_view = gtk_text_view_new();
  gtk_text_view_set_left_margin(GTK_TEXT_VIEW(text_view), 5);
  if (writeable)
  {
    gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), TRUE);
    gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text_view), TRUE);
  }
  else
  {
    gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), FALSE);
    gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text_view), FALSE);
  }
  text = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));

/* toplevel window with vbox, containing menu bar, hbox with tree and text */
/* widgets in scrolled windows, and status bar */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(window), 600, 400);
  g_signal_connect (G_OBJECT(window), "delete_event",
                    G_CALLBACK(gtk_main_quit), NULL);
  gtk_widget_show(window);
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(window), vbox);

/* menu bar with accelerator keys */
  accel = gtk_accel_group_new();
  menubar = create_menu(GTK_WINDOW(window), GTK_TREE_VIEW(gtk_tree_view),
                        text, accel);
  gtk_widget_show(menubar);
  gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);
  gtk_window_add_accel_group(GTK_WINDOW(window), accel);

/* horizontally adjustable panes with two scrolled windows: the left one for */
/* the ACPI tree view, the right one for the text view */
  pane = gtk_hpaned_new();
  gtk_paned_set_position(GTK_PANED(pane), 400);
  gtk_widget_show(pane);
  gtk_box_pack_start(GTK_BOX(vbox), pane, TRUE, TRUE, 1);

/* first scrolled window with ACPI tree view */
  scrolled_win1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scrolled_win1),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_widget_show(scrolled_win1);
  gtk_paned_add1(GTK_PANED(pane), scrolled_win1);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_win1),
                                        gtk_tree_view);

/* now fill the ACPI tree view with content: */
/* two columns for dir/file name or file entry label and file entry value.   */
/* we need a text renderer to create a display column in the Gtk tree view.  */
/* the text renderer expects the text in the 1st (dir/file name or label of  */
/* file enty) and 2nd (file entries: value or " ", dir/files: ""), and the   */
/* color in 2nd data column of the Gtk tree view. columns are made resizable.*/
  renderer1 = gtk_cell_renderer_text_new();
  column1 = gtk_tree_view_column_new_with_attributes(PROC, renderer1,
                                                     "text", COL_LABEL,
                                                     "foreground", COL_COLOR,
                                                     NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(gtk_tree_view), column1);
  gtk_tree_view_column_set_resizable(column1, TRUE);

  renderer2 = gtk_cell_renderer_text_new();
  column2 = gtk_tree_view_column_new_with_attributes("", renderer2,
                                                     "markup", COL_VALUE,
                                                     "foreground", COL_COLOR,
                                                      NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(gtk_tree_view), column2);
  gtk_tree_view_column_set_resizable(column2, TRUE);

  g_signal_connect_after(G_OBJECT (select), "changed",
			 G_CALLBACK(on_selection_changed), text);
  g_signal_connect(G_OBJECT(gtk_tree_view), "row-activated",
                   G_CALLBACK(on_activated), text);

/* second scrolled window for text view */
  scrolled_win2 = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(scrolled_win2);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_win2),
				 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_AUTOMATIC);
  gtk_paned_add2(GTK_PANED(pane), scrolled_win2);

/* text drawing styles ("tags" in gtk speech) */
  gtk_text_buffer_create_tag(text, "italic", "style", PANGO_STYLE_ITALIC, NULL);
  gtk_text_buffer_create_tag(text, "bold", "weight", PANGO_WEIGHT_BOLD, NULL);  
  gtk_text_buffer_create_tag(text, "monospace", "family", "monospace", NULL);
  gtk_text_buffer_create_tag(text, "center", "justification",
                             GTK_JUSTIFY_CENTER, NULL);

/* and the start text itself */
  gtk_text_buffer_get_iter_at_offset(text, &iter, 0);
  gtk_text_buffer_insert_with_tags_by_name(text, &iter, "\n  ", -1, "bold", 0);
  gtk_text_buffer_insert_with_tags_by_name(text, &iter, VERSION, -1, "bold", 0);
  gtk_text_buffer_insert(text, &iter, "\n  (C) 2002 Oliver Diedrich", -1);
  gtk_widget_show(text_view);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_win2),
                                        text_view);

/* status bar with message */
  statusbar = gtk_statusbar_new();  
  gtk_widget_show(statusbar);
  gtk_box_pack_start(GTK_BOX(vbox), statusbar, FALSE, FALSE, 0);

  status_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "text");
  if (update) sprintf(str, " update interval %d seconds", update);
  else sprintf(str, " static view of tree ('-u 0' on command line)");
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), status_id, str);

/* only for debugging purposes: get tree model from view and print its rows
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  gint row_count = 0;
  gboolean valid;
  
  model = gtk_tree_view_get_model(gtk_tree_view);
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while (valid)
  {
    gchar *str_data;

    gtk_tree_model_get (model, &iter, 0, &str_data, -1);
    g_print ("Row %d: %s\n", row_count, str_data);
    g_free (str_data);
    row_count ++;
    valid = gtk_tree_model_iter_next (model, &iter);
  }
}
*/

/* set up a timer to update the tree */
  if (sig && update)          /* use UNIX signals */
  {
    timer.it_interval.tv_sec = update;
    timer.it_interval.tv_usec = 0;
    timer.it_value.tv_sec = update;
    timer.it_value.tv_usec = 0;
    if (-1 == setitimer(ITIMER_REAL, &timer, NULL))
    {
      fprintf(stderr, "can't set timer for update signals\n");
      return 1;
    }
    signal(SIGALRM, update_handler);
  }
  else
  {
    printf(">>> Without using Unix signals\n\n");
    if (update) gtk_timeout_add(update*1000, (GtkFunction) update_handler, 0);
  }

  gtk_main();
  return 0;
}
