/*
 *	DMI decode rev 1.2
 *
 *	(C) 2000,2001 Alan Cox <alan@redhat.com>
 *
 *      reads some info from /proc files (C) 2002 Oliver Diedrich
 *      
 *	Licensed under the GNU Public license. If you want to use it in with
 *	another license just ask.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/utsname.h>
#include "acpiw.h"


typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;


struct dmi_header
{
	u8	type;
	u8	length;
	u16	handle;
};

static char *dmi_string(struct dmi_header *dm, u8 s)
{
	u8 *bp=(u8 *)dm;
	if (!s) return "";
	
	bp+=dm->length;
	while(s>1)
	{
		bp+=strlen(bp);
		bp++;
		s--;
	}
	return bp;
}



char *mkstr(char str[])
/* allocates memory for string str, copies str to it, and removes '\n' and */
/* blanks at end */
/* NULL: error, else: copied string */
{
  char *c;
  int l = strlen(str);
  c = malloc(l+1);
  if (!c) return NULL;
  memset(c, 0, l+1);
  strncpy(c, str, l);
  l--;
  while ( ('\n' == c[l]) || (' ' == c[l]) )
  {
    c[l] = '\0';
    l--;
    if (0 == l) break;
  }
  return c;
} /* mkstr() */



static char *dmi_chassis_type(u8 code)
{
	static char *chassis_type[]={
		"",
		"Other",
		"Unknown",
		"Desktop",
		"Low Profile Desktop",
		"Pizza Box",
		"Mini Tower",
		"Tower",
		"Portable",
		"Laptop",
		"Notebook",
		"Hand Held",
		"Docking Station",
		"All in One",
		"Sub Notebook",
		"Space-saving",
		"Lunch Box",
		"Main Server Chassis",
		"Expansion Chassis",
		"SubChassis",
		"Bus Expansion Chassis",
		"Peripheral Chassis",
		"RAID Chassis",
		"Rack Mount Chassis",
		"Sealed-case PC",
	};
	code &= ~0x80;
	
	if(code>0x18)
		return "";
	return chassis_type[code];
	
}



static char *dmi_processor_family(u8 code)
{
	static char *processor_family[]={
		"",
		"Other",
		"Unknown",
		"8086",
		"80286",
		"Intel386 processor",
		"Intel486 processor",
		"8087",
		"80287",
		"80387",
		"80487",
		"Pentium processor Family",
		"Pentium Pro processor",
		"Pentium II processor",
		"Pentium processor with MMX technology",
		"Celeron processor",
		"Pentium II Xeon processor",
		"Pentium III processor",
		"M1 Family",
		"M1","M1","M1","M1","M1","M1", /* 13h - 18h */
		"K5 Family",
		"K5","K5","K5","K5","K5","K5", /* 1Ah - 1Fh */
		"Power PC Family",
		"Power PC 601",
		"Power PC 603",
		"Power PC 603+",
		"Power PC 604",
	};
	
	if(code == 0xFF)
		return "Other";
	
	if (code > 0x24)
		return "";
	return processor_family[code];
}



static void dmi_table(int fd, u32 base, int len, int num, sysinfo *si)
{
	char *buf=malloc(len);
	struct dmi_header *dm;
	u8 *data;
	int i=0;
	
	if(lseek(fd, (long)base, 0)==-1)
	{
		perror("dmi: lseek");
		return;
	}
	if(read(fd, buf, len)!=len)
	{
		perror("dmi: read");
		return;
	}
	data = buf;
	while(i<num)
	{
		dm=(struct dmi_header *)data;
		
		switch(dm->type)
		{
		case  0:
			si->bios_vendor = mkstr(dmi_string(dm, data[4]));
			si->bios_version = mkstr(dmi_string(dm, data[5]));
			si->bios_release = mkstr(dmi_string(dm, data[8]));
			break;
			
		case 1:
			si->sys_vendor = mkstr(dmi_string(dm, data[4]));
			si->sys_product = mkstr(dmi_string(dm, data[5]));
			si->sys_serial = mkstr(dmi_string(dm, data[7]));
			break;
			
		case 2:
			si->board_vendor = mkstr(dmi_string(dm, data[4]));
			si->board_product = mkstr(dmi_string(dm, data[5]));
			si->board_version = mkstr(dmi_string(dm, data[6]));
			si->board_serial = mkstr(dmi_string(dm, data[7]));
			break;
			
		case 3:
			si->chassis = mkstr(dmi_chassis_type(data[5]));
			break;
			
		case 4:
			si->cpu_maker = mkstr(dmi_string(dm, data[7]));
			si->cpu_family = mkstr(dmi_processor_family(data[6]));
			si->cpu_version = mkstr(dmi_string(dm, data[0x10]));
			break;
			
		}
		data+=dm->length;
		while(*data || data[1])
			data++;
		data+=2;
		i++;
	}
	free(buf);
}



int get_dmi_info(sysinfo *si)
/* reads some of the DMI info and puts it into si */
/* 0: error, else : ok */
{
	unsigned char buf[20];
	int fd=open("/dev/mem", O_RDONLY);
	long fp=0xE0000L;

	si->sys_vendor = NULL;
	si->sys_product = NULL;
	si->sys_serial = NULL;
	si->bios_vendor = NULL;
	si->bios_version = NULL;
	si->bios_release = NULL;
	si->board_vendor = NULL;
	si->board_product = NULL;
	si->board_version = NULL;
	si->board_serial = NULL;
	si->cpu_maker = NULL;
	si->cpu_family = NULL;
	si->cpu_version = NULL;
	si->chassis = NULL;

	if(fd==-1)
	{
        	perror("/dev/mem");
		return 0;
	}
	if(lseek(fd,fp,0)==-1)
	{
		perror("seek");
		return 0;
	}
	
	fp -= 16;
	
	while( fp < 0xFFFFF)
	{
		fp+=16;
		if(read(fd, buf, 16)!=16)
                {
			perror("read");
                        return 0;
                }
		if(memcmp(buf, "_DMI_", 5)==0)
		{
			u16 num=buf[13]<<8|buf[12];
			u16 len=buf[7]<<8|buf[6];
			u32 base=buf[11]<<24|buf[10]<<16|buf[9]<<8|buf[8];
			dmi_table(fd, base,len, num, si);
		}
	}
	close(fd);
	return 1;
} /* get_dmi_info() */




void get_basic_sysinfo(sysinfo *si)
/* reads some sys info from /proc/cpuinfo and puts it into si */
/* A value of 0 or NULL means "could not be read" */
{
  FILE *f;
  char str[81];
  char *c;
  struct stat st;
  struct utsname kernel;

  si->cpu = NULL;
  si->mhz = 0;
  si->mem = 0;
  f = fopen("/proc/cpuinfo", "r");
  if (f)
  {
    while (fgets(str, 80, f))
    {
      if (0 == strncmp(str, "model name", 9))
      {
        c = str;
        while ( (*c != ':') && (*c != '\0') ) c++;
        if (':' == *c)
        {
          c++;
          while ( (' ' == *c) && (*c != '\0') ) c++;
          if (*c) si->cpu = mkstr(c);
        }
      }
      else if (0 == strncmp(str, "cpu MHz", 7))
      {
        c = str;
        while ( (*c != ':') && (*c != '\0') ) c++;
        if (':' == *c)
        {
          c++;
          while ( (!isdigit(*c)) && (*c != '\0') ) c++;
          if (*c) si->mhz = atoi(c);
        }
      }
    } /* while fgets() */
    fclose(f);
  } /* if (f) */
  if (stat("/proc/kcore", &st) != -1) si->mem = st.st_size;
  if (uname(&kernel) != -1)
  {
    si->sysname = mkstr(kernel.sysname);
    si->release = mkstr(kernel.release);
    si->version = mkstr(kernel.version);
    si->machine = mkstr(kernel.machine);
  }
} /* get_basic_sysinfo() */



