/*
  text.c: text mode interface for acpiw
  (C) 2002 Oliver Diedrich

  This file may be redistributed under the terms of the
  GNU Public License (GPL), see www.gnu.org
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#include "acpiw.h"


int got_signal = 0;      /* if != 0, user pressed Ctrl-C */




void sighandler(int sig)
/* catches signals SIGHUP, SIGINT und SIGTERM and sets got_signal */
{
  got_signal = sig;
}




void  hexdump(unsigned char *buf)
/* dumps the binary data in buf to stdout. Attention: Bytes 0-3 in buf are */
/* garbage, bytes 4-7 contain the length of the buffer including the eight */
/* leading bytes */
{
  int *size = (int *) (buf+4);     /* size of buf is stored at offset 4 */
  int i, j;
  char str[80];                    /* not really necessary, we could write */
  char *c = str;                   /* directly to stdout; but it works ... */
  int n = 0;                       /* bytes written */

  buf += 8;
  *size -= 8;

  for (i=0; i<((*size)/16)+1; i++)
  {
    c = str;
    sprintf(c, "%05X  ", i*16);
    c += 7;
    for (j=0; j<16; j++, buf++, n++)    /* hex display */
    {
      if (8 == j)
      {
        sprintf(c, " ");
        c++;
      }
      if (n<*size) sprintf(c, "%02X ", *buf);
      else sprintf(c, "   ");
      c += 3;
    }
    buf -= j;
    n -= j;
    sprintf(c, " ");
    c++;
    for (j=0; j<16; j++, buf++, n++)    /* character display */
    {
      if (8 == j)
      {
        sprintf(c, " ");
        c++;
      }
      if (n<*size) sprintf(c, "%c", ((*buf < 32) || (*buf > 127)) ? '.' : *buf);
      c++;
    }
    printf("%s\n", str);
  } /* for i */
} /* hexdump() */




void print_tree(file *tree, int depth, int changed)
/* prints tree starting with tree, indented depth*2 blanks */
/* if (changed) then only changed entries are printed without indentation */
{
  file *f;
  line *l;
  char fill[3] = "  \0";
  int i;

assert(tree != NULL);
assert(tree->name != NULL);

  f = tree;
  while(f)
  {
    if (changed && !f->changed)    /* print only changed entries, but this */
    {                              /* file/directory has not changed */
      f = f->next;
      continue;
    }

    if (changed)
    {
      if (S_ISDIR(f->mode)) printf("%s/", f->name);
      else printf("%s\n", f->name);
    }
    else
    {
      for (i=0; i<depth; i++) printf("%s", fill);
      printf("+-%s\n", f->name);
    }
    
    if (S_ISDIR(f->mode) && f->data) print_tree(f->data, depth+1, changed);
    else
    {

if (changed) assert(f->data);

      if (f->data)
      {
        l = f->data;
        while (l)
        {
          if (changed && !l->changed)   /* print only changed entries, but */
          {                             /* this line hasn't changed */
            l = l->next;
            continue;
          }
          if (!changed)
          {
            for (i=0; i<depth+1; i++) printf("%s", fill);
          }
          printf("%s%s", fill, l->latest->label);
          if (l->latest->value)
          {
            if (strlen(l->latest->label) < 25)
            {
              for (i=0; i<25-(strlen(l->latest->label)); i++) printf(" ");
            }
            if (!strncmp(l->label, "/proc/", 6))  /* a *dt file */
            {
              printf(":\n");
              hexdump(l->value);
            }
            else printf(": %s\n", l->latest->value);
          }
          else printf("\n");
          l = l->next;
        }
      }
    } /* else */
    f = f->next;
  } /* while(e) */
} /* print_tree() */




int main(int argc, char **argv)
{
  file *acpi_tree;       /* complete /proc/acpi tree at startup */
  int i;
  int update = 5;        /* update interval in seconds (-u) */
  time_t seconds;        /* time in seconds after the epoch */
  struct timeval tv;     /* needed for gettimeofday() */
  char *fname = NULL;    /* file to save data in */
  char **acpi_info;      /* ACPI info read from syslog */
  int n_acpi_info;       /* # lines in acpi_info */
  char acpi_version[10]; /* ACPI version from /proc/acpi/info */
  sysinfo si;            /* system  information */
  char **c;

  printf("%s\n", VERSION);
  printf("(C) 2002 Oliver Diedrich\n\n");

  for (i=1; i<argc; i++)
  {
    char *c = argv[i];
    if (*c != '-') usage(argv[0]);
    c++;
    switch(*c)
    {
      case 'u': if (i==argc-1) usage(argv[0]);    /* last argument */
                i++;
                update = strtol(argv[i], NULL, 10);
                break;

      case 's': if (i==argc-1) usage(argv[0]);    /* last argument */
                i++;
                fname = argv[i];
                break;

      default:  usage(argv[0]);
    }
  }

  signal(SIGINT, sighandler);
  signal(SIGHUP, sighandler);
  signal(SIGTERM, sighandler);
  if (-1 == gettimeofday(&tv, NULL)) seconds = 0;
  else seconds = tv.tv_sec;

  get_basic_sysinfo(&si);
  get_dmi_info(&si);
  get_acpi_version(PROC, acpi_version);

  printf("System information\n\n");
  printf("System:\t%s %s (%s); serial #: %s\n", 
             si.sys_vendor, si.sys_product,
             si.chassis, si.sys_serial);
  printf("BIOS:\t%s %s (%s)\n",
             si.bios_vendor, si.bios_version, si.bios_release);
  printf("Board:\t%s %s (%s); serial #: %s\n",
              si.board_vendor, si.board_product,
              si.board_version, si.board_serial);
  printf("CPU:\t%s %s, %d MHz\n", si.cpu_maker, si.cpu, si.mhz);
  printf("RAM:\t%d MByte\n", si.mem / (1024*1023));
  printf("Kernel:\t%s %s for %s (build %s)\n",
             si.sysname, si.release, si.machine, si.version);
  printf("ACPI:\trelease %s\n", acpi_version ? acpi_version : "");

  n_acpi_info = get_acpi_info(&acpi_info);
  c = acpi_info;
  printf("\nACPI related syslog messages:\n\n");
  for (i=0; i<n_acpi_info; i++, c++) printf("%s\n", *c);
  printf("\n");

  acpi_tree = build_tree(PROC);
  if (acpi_tree)
  {
    printf("\n%s at %s", PROC, ctime(&seconds));
    print_tree(acpi_tree, 0, 0);
    fflush(stdout);
  }
  else
  {
    fprintf(stderr, "\nerror while reading %s\n", PROC);
    return 1;
  }

  if (!update) return 0;

  while(1)
  {
    if (got_signal)      /* user pressed Ctrl-C */
    {
      if (fname)
      {
        FILE *f = fopen(fname, "w");
        if (!f) fprintf(stderr, "\ncan't open file '%s'\n", fname);
        else
        {
          if (save_all(acpi_tree, f, PROC, acpi_info, n_acpi_info,
                       acpi_version, si))
           printf("\nACPI data written to '%s'\n", fname);
          else
            printf("\nerror while wrinting ACPI data to %s\n", fname);
          fclose(f);
        }
      }
      else printf("\nNo '-s' given, ACPI data are not saved.\n");
      return 0;
    }
        
    sleep(update);
    if (-1 == gettimeofday(&tv, NULL)) seconds = 0;
    else seconds = tv.tv_sec;
    if (rescan_tree(acpi_tree, PROC))
    {
      printf("\nChanged entries at %s", ctime(&seconds));
      print_tree(acpi_tree, 0, 1);
      fflush(stdout);
    }
    else
    {
      printf("\nno changes\n");
      fflush(stdout);
    }
  }

  return 0;
} /* main() */
